/*
 * Hash abstract data type.
 * Hash.c
 */

#define __HASH_C

#include "hash.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>


/* DEFINITION DES TYPES DE DONNEE */

/* cellule */
typedef struct Cell *Cell;

typedef struct Cell {
  Cell next;
  char *ident;
  int data_size;
  void *data;
} _Cell;

/* table de hashage */
typedef struct Hash {
  int hash_size;
  int ident_size;
  Cell *table;
} _Hash;

static char _magic_word[] = "HashTable\n";

/* FONCTION LOCALES */

/* Calcule l'index dans la table de la liste de cellules
 * pour l'identificateur donne
 */
static int _calc_index(int hash_size, char *ident) {
  int somme = 0;

  /* Vrification des arguments */
  assert(hash_size);
  assert(ident);

  /* Ajouter les valeurs ASCII de tous les
   * caracteres de l'identificateur
   */
  while(*ident) {
    somme += (int) *ident;
    ident++;
  }
  /* Retourner la somme des valeurs ASCII modulo
   * la taille de la table
   */
  return(somme % hash_size);
}

/* Fonction de rappel de la fonction hash_all.
 * Sauve l'entree de la table a la fin du fichier
 * passe en parametre libre.
 */ 
static int _save_cells(void *param, char *ident, void *data, int data_size) {
  int status;
  FILE *file = (FILE*)param; /* recuperation du fichier */
  int ident_size = strlen(ident) + 1;

  /* Ecriture de la taille de l'identificateur */
  status = fwrite(&ident_size, sizeof(int), 1, file);
  if(status!=1) return(0);
  /* Ecriture de l'identificateur */
  status = fwrite(ident, ident_size, 1, file);
  if(status!=1) return(0);

  /* Ecriture de la taille de la donnee */
  status = fwrite(&data_size, sizeof(int), 1, file);
  if(status!=1) return(0);
  /* Ecriture de la donnee */
  status = fwrite(data, data_size, 1, file);
  if(status!=1) return(0);

  return(1);
}

/* Lecture d'une cellule.
 */

static int _load_cells(FILE *file, Hash hash) {
  int ident_size;
  int data_size;
  char *ident = 0;
  void *data = 0;
  int error = 0;

  /* Lecture de la taille de l'identificateur */
  if(fread(&ident_size, sizeof(int), 1, file)!=1) {
    goto Error;
  }
  /* Allocation pour l'identificateur */
  ident = malloc(ident_size);
  if(!ident) {
    goto Error;
  }
  /* Lecture de l'identificateur */
  if(fread(ident, ident_size, 1, file)!=1) {
    goto Error;
  }

  /* Lecture de la taille de la donnee */
  if(fread(&data_size, sizeof(int), 1, file)!=1) {
    goto Error;
  }
  /* Allocation pour la donnee */
  data = malloc(data_size);
  if(!data) {
    goto Error;
  }
  /* Lecture de la donnee */
  if(fread(data, data_size, 1, file)!=1) {
    goto Error;
  }

  /* Ajout dans la table du couple
   * identificateur-donnee */
  hash_add(hash, ident, data, data_size);

  error = 1;

 Error:
  /* Liberation memoire */
  if(ident) free(ident);
  if(data) free(data);

  return(error);
}


/* FONCTIONS PUBLIQUES */

/* Creation d'une hash table vide */
Hash hash_new(int hash_size, int ident_size) {
  Hash hash;

  /* Verification des arguments */
  assert(ident_size > 0);
  assert(hash_size > 0);

  /* Creation de la table */
  hash = malloc(sizeof(_Hash));
  if(!hash) return(0);

  /* Initialisation de la table */
  hash->hash_size = hash_size;
  hash->ident_size = ident_size;
  hash->table = calloc(hash_size, sizeof(Cell));
  if(!hash->table) {
    free(hash);
    return(0);
  }

  return(hash);
}

/* Destruction d'une table */
HashStatus hash_destroy(Hash hash) {
  /* Vrification des arguments */
  assert(hash);

  /* Desallocation de la table */
  free(hash->table);
  free(hash);

  return(HASH_OK);
}

/* Ajoute une entree dans la hash table. La taille
 * de la donne est variable. La longueur de l'identificateur
 * est fixee  la creation de la table
 */
HashStatus hash_add(Hash hash, char *ident, void *data, int data_size) {
  Cell cell;
  int index;

  /* Verifications des arguments .
   * Actif que lorsque la macro NDEBUG n'est pas definie
   */
  assert(hash);
  assert(ident);
  assert(strlen(ident) <= hash->ident_size);
  assert(data);
  assert(data_size > 0);

  if(hash_size(hash, ident) > 0)
    return(HASH_DOUBLON);

  /* Creation d'une cellule. */
  cell = malloc(sizeof(_Cell) + hash->ident_size + 1 + data_size);
  if(!cell) return(HASH_ERROR);

  /* Initialisaton de la cellule.
   * Le format des cellule est le suivant:
   *
   *   CELLULE
   *   +-----------------+
   *   | suivante -------|----->...
   *   | identificateur -|--
   *   | taille_donnee   |  |
   *   | donnee ---------|--|--
   *   |.................|  |  |
   *   |                 |<-   |
   *   | IDENTIFICATEUR  |     |
   *   |                 |     |
   *   |.................|     |
   *   |                 |<----
   *   |                 |
   *   | DONNEE          |
   *   |                 |
   *   |                 |
   *   +-----------------+
   *
   */
  cell->ident = ((char *)cell) + sizeof(_Cell);
  cell->data_size = data_size;
  cell->data = ((char *)cell) + sizeof(_Cell) + hash->ident_size + 1;
  strcpy(cell->ident, ident);
  memcpy(cell->data, data, data_size);

  /* Ajout de la cellule dans la table. */
  index = _calc_index(hash->hash_size, ident);
  cell->next = hash->table[index];
  if(hash->table[index])
    hash->table[index]->next = cell;
  else
    hash->table[index] = cell;

  return(HASH_OK);
}

/* Retourne la taille d'une cellule */
int hash_size(Hash hash, char *ident) {
  Cell cell;
  int index;

  /* Verification des arguments */
  assert(hash);
  assert(ident);

  /* Calcule de l'index */
  index = _calc_index(hash->hash_size, ident);
  cell = hash->table[index];

  /* Pour toutes les cellules de la liste */
  while(cell) {
    /* Si les identificateurs sont identiques,
     * bingo! 
     */
    if(!strcmp(ident, cell->ident)) {
      return(cell->data_size);
    }
    /* Sinon, continuer */
    cell = cell->next;
  }
  return(-1);
}

/* Retourne le contenu d'une cellule */
HashStatus hash_data(Hash hash, char *ident, void *data, int *data_size) {
  Cell cell;
  int index;

  /* Verification des arguments */
  assert(hash);
  assert(ident);

  /* Calcul de l'index */
  index = _calc_index(hash->hash_size, ident);
  cell = hash->table[index];

  /* Pour toutes les cellules */
  while(cell) {
    /* Si les identificateurs sont identiques,
     * Jackpot!
     */
    if(!strcmp(ident, cell->ident)) {
      /* Retourner  l'utilisateur ce qui desire */
      if(data_size) {
	*data_size = cell->data_size;
      }
      if(data) {
	memcpy(data, cell->data, cell->data_size);
      }
      return(HASH_OK);
    }
    /* Sinon, continuer */
    cell = cell->next;
  }
  return(HASH_NOFOUND);
}

/* Supression d'une cellule */
HashStatus hash_supp(Hash hash,	char *ident) {
  Cell cell, prev;
  int index;

  /* Verification des arguments */
  assert(hash);
  assert(ident);

  /* Calcule de l'index */
  index = _calc_index(hash->hash_size, ident);
  
  /* Parcourt de la liste de cellules */
  cell = hash->table[index];
  prev = 0;
  while(cell) {
    /* Comparaison des identificateurs */
    if(!strcmp(ident, cell->ident)) {
      /* Mise  jour des liens */
      if(prev) {
	prev->next = cell->next;
      }
      else {
	hash->table[index] = cell->next;
      }
      /* Destruction de la callule */
      free(cell);
      return(HASH_OK);
    }
    /* Sinon, continuer */
    prev = cell;
    cell = cell->next;
  }
  return(HASH_NOFOUND);
}

/* Applique  toutes les cellules une fonction */
HashStatus hash_all(Hash hash, void *param,
		    int (* proc)(void *param, char *ident, void *data, int data_size)) {
  int index;

  /* Verification des arguments */
  assert(hash);
  assert(proc);

  /* Pour toutes les listes de la table */
  for(index = 0; index < hash->hash_size; index++) {
    Cell cell = hash->table[index];

    /* Pour toutes les cellules de la liste */
    while(cell) {
      /* Appliquer la fonction. Par convention, si
       * elle retourne zero, interrompre le parcourt
       */
      if(!proc(param, cell->ident, cell->data, cell->data_size)) {
	return(HASH_ERROR);
      }
      /* Sinon, continuer */
      cell = cell->next;
    }
  }
  return(HASH_OK);
}

/* Sauve une hash table dans un fichier */
HashStatus hash_save(Hash hash, char *filename) {
  FILE *file;
  HashStatus status = HASH_ERROR;

  /* Verification des arguments */
  assert(hash);
  assert(filename && *filename);

  /* Ouverture du fichier en ecriture. Le contenu
   * precedent, s'il existait, est efface */
  file = fopen(filename, "w");
  if(!file) {
    goto Error;
  }

  /* Ecriture du mot magique */
  if(fwrite(_magic_word, strlen(_magic_word), 1, file)!=1) {
    goto Error;
  }

  /* Ecriture de la taille de la table 
   * et de la taille de l'identificateur */
  if(fwrite(&hash->hash_size, sizeof(int), 1, file)!=1) {
    goto Error;
  }
  if(fwrite(&hash->ident_size, sizeof(int), 1, file)!=1) {
    goto Error;
  }

  /* La fonction de rappel _save_cell
   * est appelee pour toutes les entrees
   * de la table */
  status = hash_all(hash, (void*)file, _save_cells);

 Error:
  if(file) fclose(file);
  if(status!=HASH_OK) {
    /* En cas d'erreur, effacer le fichier */
    unlink(filename);
  }

  return(status);
}

/* Charge une hash table a partir d'un fichier */
Hash hash_load(char *filename) {
  int hash_size;
  int ident_size;
  char tampon[100];
  Hash hash = 0;
  FILE *file;

  /* Verification des arguments */
  assert(filename && *filename);

  /* Ouverture du fichier en lecture */
  file = fopen(filename, "r");
  if(!file) {
    goto Error;
  }
  
  /* Lecture du mot magique et verification*/
  if(fread(tampon, strlen(_magic_word), 1, file)!=1) {
    goto Error;
  }
  if(strcmp(tampon, _magic_word)) {
    goto Error;
  }

  /* Lecture de la taille de la table 
   * et de la taille de l'identificateur */
  if(fread(&hash_size, sizeof(int), 1, file)!=1) {
    goto Error;
  }
  if(fread(&ident_size, sizeof(int), 1, file)!=1) {
    goto Error;
  }

  /* Creation de la table */
  hash = hash_new(hash_size, ident_size);
  if(!hash) {
    goto Error;
  }
  /* Lecture des enregistrements 
   * jusqu'a la fin du fichier */
  { /* hack pour tester la fin de fichier (feof ?) */
    char c; 
    while((c=fgetc(file))!=EOF) {
      ungetc(c, file); /* tordu ! non ? */
      if(!_load_cells(file, hash)) {
	goto Error;
      }
    }
  }
  fclose(file);
  return(hash);

 Error:
  /* en cas d'erreur, fermer le fichier 
   * et detruire la table */
  if(file) fclose(file);
  if(hash) hash_destroy(hash);

  return(0);
}
